(* type node *)
type node = (int * char option)
(* type btree *)
type 'a btree = 'a BinTree.t = Empty | Node of ('a btree * 'a * 'a btree)
(* type t *)
type t = node btree
(* type bits *)
type bits = int list
(* type freq *)
type freq = (char * int)Lst.t
(* type table *)
type table = (char * bits)Lst.t
(* sort : ('a -> 'a -> int) -> 'a list -> 'a list *)
let sort c xs = BinTree.flatten(BinTree.search_tree c xs)
(* collate : 'a list -> ('a * int) list *)
let rec collate = function
 | []          -> []
 | w::ws as xs ->
  let (ys,zs) = Lst.span ((=)w) xs in
  (w,Lst.length ys) :: collate zs
(* sample : Strng.t -> freq *)
let sample xs = sort (fun (c,v) (d,w) ->
 compare (v,c) (w,d)) (collate(sort compare xs))
(* weight : t -> int *)
let weight = function
 | Node(_,(w,_),_) -> w
 | _               -> failwith "empty tree"
(* insert : t -> t list *)
let insert vt wts = 
 let (xts,yts) =
  Lst.span (fun x -> weight x <= weight vt) wts in
 xts@(vt::yts)
(* combine : t list -> t list *)
let combine = function
 | xt::yt::xts -> let w = weight xt + weight yt in
  insert (Node(xt,(w,None),yt)) xts
 | _           ->
  failwith "length has to be greater than 0"
(* singleton : 'a list -> bool *)
let singleton xs = Lst.length xs = 1
(* mknode : (char * int) -> t *)
let mknode (c,w) = Node(Empty,(w,Some c),Empty)
(* tree : Strng.t -> t *)
let tree xs =
 Lst.hd(Lst.until singleton combine (Lst.map mknode xs))
(* table : t -> table *)
let rec table = function
 | Node(Empty,(_,Some c),Empty) -> [(c,[])]
 | Node(l,_,r)                  ->
   (Lst.map (fun (c,code) -> (c,0::code)) (table l))@
   (Lst.map (fun (c,code) -> (c,1::code)) (table r))
 | _ -> failwith "the Huffman tree is empty"
(* lookup : table -> char -> bits *)
let rec lookup xbs v = match xbs with
 | ((x,bs)::xbs) -> if x = v then bs else lookup xbs v
 | _             -> failwith "not found"
(* encode : table -> Strng.t -> bits *)
let encode t text = Lst.concat(Lst.map (lookup t) text)
(* decode_char : (t * bits) -> (char * bits) *)
let rec decode_char = function
 | (Node(Empty,(_,Some c),Empty),cs) -> (c,cs)
 | (Node(xt,_,_),0::cs)              -> decode_char (xt,cs)
 | (Node(_,_,xt),1::cs)              -> decode_char (xt,cs)
 | _                                 -> failwith "empty tree"
(* decode : t -> bits -> Strng.t *)
let rec decode t = function
 | [] -> []
 | xs -> let (c,xs) = decode_char (t,xs) in c::decode t xs
(* bits2int : bits -> int *)
let bits2int bs =
 let l = Lst.length bs - 1 in
 let rec bits2int i = function
  | []    -> 0
  | b::bs -> b * Int.pow2 i + bits2int (i - 1) bs
 in
 bits2int l bs
(* int2byte : int -> bits *)
let int2byte i =
 let rec int2byte acc i =
  if i = 0 then acc else int2byte (i mod 2::acc) (i / 2)
 in
 let r = int2byte [] i in
 let l = Lst.length r in
 Lst.append (Lst.init 0 (8 - l)) r
(* write_bits : bits -> out_channel -> unit *)
let write_bits bs chout =
 let rec write_bits bs count chout =
  let byte = Lst.take 8 bs in
  let len = Lst.length byte in
  if len < 8 then
   let i = bits2int (Lst.append byte (Lst.init 0 (8 - len))) in
   output_byte chout i
  else
   let i = bits2int byte in
   output_byte chout i;
   write_bits (Lst.drop 8 bs) (count + 8) chout
 in
 write_bits bs 0 chout
(* read_bits : int -> in_channel -> bits *)
let read_bits n chin =
 let rec read_bits n count chin =
  let i = input_byte chin in
  let byte = int2byte i in
  if count + 8 >= n then
   Lst.take (n - count) byte
  else
   Lst.append byte (read_bits n (count + 8) chin)
 in
 read_bits n 0 chin
(* compress : string -> string -> unit *)
let compress filein fileout =
 let i = File.read filein in
 let s = sample i in
 let tree = tree s in
 let tab = table tree in
 let e = encode tab i in
 let chout = open_out fileout in
 output_value chout tree;
 output_binary_int chout (Lst.length e);
 write_bits e chout;
 close_out chout
(* expand : string -> string -> unit *)
let expand filein fileout =
 let chin = open_in filein in
 let tree = input_value chin in
 let n = input_binary_int chin in
 let bs = read_bits n chin in
 close_in chin;
 let d = decode tree bs in
 File.write fileout d
(* toplevel_printer_freq : Format.formatter -> freq -> unit *)
let toplevel_printer_freq fmt f =
 let cmp (c,i) (d,j) = match compare i j with 0 -> compare c d
                                            | r -> r
 in
 let f = List.sort cmp f in
 Format.fprintf fmt "\n%!";
 List.iter (fun (c,i) -> Format.fprintf fmt "%c occurs %3i time(s)\n%!" c i) f
(* toplevel_printer_bits : Format.formatter -> bits -> unit *)
let toplevel_printer_bits fmt b =
 List.iter (fun i -> Format.fprintf fmt "%i%!" i) b
(* toplevel_printer_table : Format.formatter -> table -> unit *)
let toplevel_printer_table fmt t =
 Format.fprintf fmt "\n%!";
 List.iter (fun (c,b) ->
  Format.fprintf fmt "%c -> %a\n%!" c toplevel_printer_bits b) t
(* toplevel_printer_node : Format.formatter -> node -> unit *)
let toplevel_printer_node fmt = function
 | (i,None)   -> Format.fprintf fmt "%i%!" i
 | (i,Some c) -> Format.fprintf fmt "(%c,%i)%!" c i
(* toplevel_printer : Format.formatter -> t -> unit *)
let toplevel_printer fmt t =
 Format.fprintf fmt "\n%!";
 let rec print d = function
  | Empty -> ()
  | Node (l,n,r) ->
  let e =
   if d = [] then [] else Lst.take (Lst.length d - 3) d @ ['+'; '-'; ' ']
  in
  Format.fprintf fmt "%s%a\n%!"
   (Strng.to_string e)
   toplevel_printer_node n;
  print (d @ ['|'; ' '; ' ']) l;
  if l <> Empty then
   Format.fprintf fmt "%s\n%!" (Strng.to_string (d @ ['|'; ' '; ' ']));
  print (d @ [' '; ' '; ' ']) r;
 in
 print [] t
(*  *)
